#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "avl.h"

/*  
Notes: 
- less is left, more is right 
- if left subtree is longer, blnc is positive

*/

typedef struct struct_AvlNode
{
    short blnc;
    struct struct_AvlNode *left;
    struct struct_AvlNode *right;
    void *user;
} AvlNode;

typedef struct struct_AvlHead
{
    int (*cmp_func)(void *,void *);
    AvlNode *root;
} AvlHead;

typedef struct struct_AvlStack
{
    short subtree;
    AvlNode *node;
    struct struct_AvlStack *next;
} AvlStack;
    

static int AvlDoTraverse(AvlNode *,int (*)(void *));
static void AvlBalance(AvlHead *,AvlNode *,AvlNode *,AvlNode *,int);
static void AvlPrint(AvlNode *,int);
static int AvlCheck(AvlNode *,int *);
static void AvlTrashStack(AvlStack *);
static void AvlDoDestroy(AvlNode *,void (*)(void *));

void *AvlInit(int (*func)(void *,void *))
{
    AvlHead *head= (AvlHead *) malloc(sizeof(AvlHead));

    head->root= 0;
    head->cmp_func= func;
    return (void *) head;
}


int AvlTraverse(void *vroot,int (*func)(void *))
{
    AvlHead *root= (AvlHead *) vroot;
    return AvlDoTraverse(root->root,func);
}

static int AvlDoTraverse(AvlNode *p,int (*func)(void *))
{
    if (p==0) return 1;
    if (p->left)
	if (!AvlDoTraverse(p->left,func)) return 0;
    if (!(*func)(p->user)) return 0;
    if (p->right)
	if (!AvlDoTraverse(p->right,func)) return 0;
    return 1;
}

static void AvlPrint(AvlNode *p,int num)
{
    char buff[80];
    int i;

    if (p==0) return;
    if (num==0) 
    { 
	int total=0;
	int height= AvlCheck(p,&total);
	printf(" AvlPrint: height: %d total: %d\n",height,total);
    }
    if (p->right) AvlPrint(p->right,num+1);
    for (i=0;i<2*num;i++)
	buff[i]= ' ';
    buff[i]= '\0';
    strcat(buff,(char *)p->user);
    printf("%s %d\n",buff,(int)p->blnc);
    if (p->left) AvlPrint(p->left,num+1);
}

static int AvlCheck(AvlNode *n,int *total)
{
    int left,right;

    if (n==0) return 0;
    right= AvlCheck(n->right,total);
    left= AvlCheck(n->left,total);
    (*total)++;
    if (left-right != n->blnc)
    {
	printf("ERROR: AvlCheck()\n");
	printf("  node: %s\n",n->user);
	printf("  blnc: %d left: %d right: %d\n",n->blnc,left,right);
    }
    if (right > left) return right+1;
    return left+1;
}

void *AvlInsert(void *vroot,void *user)
{
    AvlNode *n; /* new node */
    AvlNode *a; /* most recent node with +/-1 */
    AvlNode *f; /* parent of A */
    AvlNode *p; /* traversal pointer */
    AvlNode *q; /* parent of P */
    AvlNode *b; /* subtree before rotation */
    short d;     /* +1 means N is on right subtree of A */
    AvlHead *root= (AvlHead *) vroot;
    int ret;

    /* handle special case of null tree */
    if (root->root==0)
    {
	n= (AvlNode *) malloc(sizeof(AvlNode));
	root->root= n;
	n->left= 0; n->right= 0; n->blnc= 0;
	n->user= user;
	return 0;
    }

    /* locate insertions point for user */
    a= 0; q= 0; p= root->root; a= p;
    while (p!=0)
    {
	if (p->blnc != 0)
	{
	    a= p; f= q;
	}
	ret= (*root->cmp_func)(user,(void *)p->user);   
	if (ret<0)
	{
	    q= p; p= p->left;
	}
	else if (ret>0)
	{
	    q= p; p= p->right;
	}
	else
	    /* node exists */
	    return p->user;
    }

    /* insert and adjust balance factors */
    n= (AvlNode *) malloc(sizeof(AvlNode));
    n->user= user;
    n->left= 0; n->right= 0; n->blnc= 0;
    if (ret<0) q->left= n;
    else q->right= n;
    /* adjust balance factors on path from A to Q */
    ret= (*root->cmp_func)(user,(void *)a->user);   
    if (ret>0) 
    {
	p= a->right; b= p; d= -1;
    }
    else
    {
	p= a->left; b= p; d= 1;
    }
    while (p!=n)
    {
	ret= (*root->cmp_func)(user,(void *)p->user);   
	if (ret>0)
	{
	    p->blnc= -1; p= p->right;
	}
	else
	{
	    p->blnc= 1; p= p->left;
	}
    }

    /* rebalance tree if necessary */
    if (a->blnc==0)
    {
	a->blnc= d;
	return 0;
    }
    if (a->blnc+d==0)
    {
	a->blnc= 0;
	return 0;
    }
    AvlBalance(root,f,a,b,d);

    return 0;
}


static void AvlBalance(AvlHead *root,AvlNode *f,AvlNode *a,AvlNode *b,int d)
{
    AvlNode *c;

    if (d==1) /* left subtree is too long */
    {
	if (b->blnc==1) /* rotation LL */
	{
	    a->left= b->right; b->right= a;
	    a->blnc= 0; b->blnc= 0;
	}
	else if (b->blnc==-1)/* rotation LR */
	{
	    c= b->right;
	    b->right= c->left;
	    a->left= c->right;
	    c->left= b;
	    c->right= a;
	    switch (c->blnc)
	    {
		case -1:
		    b->blnc= 1; a->blnc= 0;
		    break;
		case 0:
		    b->blnc= 0; a->blnc= 0;
		    break;
		case 1:
		    a->blnc= -1; b->blnc= 0;
		    break;
	    }
	    c->blnc= 0;
	    b= c;
	}
	else /* special rotation encountered in deletion */
	{
	    a->left= b->right; b->right= a;
	    a->blnc= 1; b->blnc= -1;
	}
    }
    else /* right subtree is too long */
    {
	if (b->blnc==-1) /* rotation RR */
	{       
	    a->right= b->left; b->left= a;
	    a->blnc= 0; b->blnc= 0;
	}
	else if (b->blnc==1) /* rotation RL */
	{       
	    c= b->left;
	    b->left= c->right;
	    a->right= c->left;
	    c->right= b;
	    c->left= a;
	    switch (c->blnc)
	    {
		case -1:
		    a->blnc= 1; b->blnc= 0;
		    break;
		case 0:
		    b->blnc= 0; a->blnc= 0;
		    break;
		case 1:
		    b->blnc= -1; a->blnc= 0;
		    break;
	    }
	    c->blnc= 0;
	    b= c;
	}
	else
	{
	    a->right= b->left; b->left= a;
	    a->blnc= -1; b->blnc= 1;
	}
    }

    /* reset parent of rotated subtree */
    if (f==0) root->root= b;
    else if (f->left==a) f->left= b;
    else f->right= b;
}


void *AvlDelete(void *vroot,void *user)
{
    AvlHead *root= (AvlHead *) vroot;
    AvlNode *node= root->root;
    AvlNode **prnt;      /* address of deleted node */
    AvlStack *stk,*tmp,*deleted;
    void *retuser;
    int ret;

    if (!node) return 0;
    /* initialize stack */
    stk= (AvlStack *) malloc(sizeof(AvlStack));
    stk->node= node;
    stk->next= 0;

    /* find the node to be deleted */
    while (1)
    {
	ret= (*root->cmp_func)(user,(void *)node->user);        
	if (ret<0)
	{
	    if (node->left==0) /* ERROR: trash stack and exit */
	    {
		AvlTrashStack(stk);
		return 0;
	    }
	    stk->subtree= 1;
	    tmp= (AvlStack *) malloc(sizeof(AvlStack));
	    tmp->next= stk;
	    stk= tmp;
	    node= node->left;
	    stk->node= node;
	}
	else if (ret>0)
	{
	    if (node->right==0) /* ERROR: trash stack and exit */
	    {
		AvlTrashStack(stk);
		return 0;
	    }
	    stk->subtree= -1;
	    tmp= (AvlStack *) malloc(sizeof(AvlStack));
	    tmp->next= stk;
	    stk= tmp;
	    node= node->right;
	    stk->node= node;
	}
	else if (ret==0)
	{
	    deleted= stk;
	    retuser= node->user;
	    /* set prnt */
	    if (stk->next==0)
		prnt= &(root->root);
	    else if (stk->next->subtree==-1)
		prnt= &(stk->next->node->right);
	    else
		prnt= &(stk->next->node->left);
	    break;
	}
    }

    /* stack is at the deleted node */
    /* begin deletion algorithm in 3 cases */
    if (node->right==0) 
    {
	*prnt= node->left;
	free(node);
	tmp= stk->next; /* pop the stack once */
	free(stk);
	stk= tmp;
    }
    else if (node->right->left==0)
    {
	*prnt= node->right;
	node->right->left= node->left;
	node->right->blnc= node->blnc;
	stk->node= node->right;
	stk->subtree= -1;
	free(node);
    }
    else /* find successor of node and replace node with it */
    {
	AvlNode *del;

	stk->subtree= -1;
	tmp= (AvlStack *) malloc(sizeof(AvlStack));
	tmp->next= stk;
	stk= tmp;
	node= node->right;
	stk->node= node;
	stk->subtree= 1;
	while (node->left->left!=0)
	{
	    tmp= (AvlStack *) malloc(sizeof(AvlStack));
	    tmp->next= stk;
	    stk= tmp;
	    node= node->left;
	    stk->node= node;
	    stk->subtree= 1;
	}
	del= node->left;
	node->left= del->right;
	deleted->node->user= del->user;
	free(del);
    }

    /* rebalance tree starting with stack */
    /* note that 'node' is invalid */
    while (stk!=0)
    {
	node= stk->node;
	if (node->blnc==stk->subtree)
	{
	    node->blnc= 0;
	    tmp= stk;
	    stk= stk->next;
	    free(tmp);
	}
	else if (node->blnc==0)
	{
	    node->blnc= -stk->subtree;
	    AvlTrashStack(stk); /* terminate algorithm */
	    stk= 0;
	}
	else 
	{
	    /* AvlBalance(root,f,a,b,d); */
	    AvlNode *f,*b;
	    int terminate= 0;
	    if (stk->next==0) f= 0;
	    else f= stk->next->node;
	    if (stk->subtree==-1) b= node->left;
	    else b= node->right;
	    if (b->blnc==0) terminate= 1;
	    AvlBalance(root,f,node,b,-stk->subtree);
	    if (terminate)
	    {
		AvlTrashStack(stk); /* terminate algorithm */
		stk= 0;
	    }
	    else
	    {
		tmp= stk;
		stk= stk->next;
		free(tmp);
	    }
	}       
    }
    return retuser;
}

static void AvlTrashStack(AvlStack *stk)
{
    AvlStack *tmp;

    while (stk!=0) 
    {
	tmp= stk;
	stk= stk->next;
	free(tmp);
    }
}

void *AvlSearch(void *vroot,void *user)
{
    AvlHead *root= (AvlHead *) vroot;
    AvlNode *n= root->root;
    int ret;

    while (n)
    {
	ret= (*root->cmp_func)(user,(void *)n->user);   
	if (ret==0) return n->user;
	if (ret<0) n= n->left;
	else n= n->right;
    }
    return 0;
}

void AvlDestroy(void *vroot,void (*func)(void *))
{
    AvlHead *root= (AvlHead *) vroot;

    AvlDoDestroy(root->root,func);
    free(root);

}

void AvlDoDestroy(AvlNode *n,void (*func)(void *))
{
    if (!n) return;
    if (func) (*func)(n->user);
    if (n->right) AvlDoDestroy(n->right,func);
    if (n->left) AvlDoDestroy(n->left,func);
    free(n);
}

